/*---------------------------------------------------------------------------
 bsvm2_mex.c: MEX-file for multi-class B-SVM with L2-soft margin.

 Compile: 
  mex -O -I../../kernels -outdir ../ bsvm2_mex.c ../../kernels/kernel_fun.c npa.c kozinec.c mdm.c

 Synopsis:
  [Alpha,bias,exitflag,kercnt,trnerr,t,UB,LB,History] = 
    bsvm2_mex(data,labels,ker,arg,C,solver,tmax,tolabs,tolrel)

 Input:
  data [dim x num_data] Training vectors.
  labels [1 x num_data] Labels.
  ker [string] Kernel identifier.
  arg [1 x nargs] Kernel argument.
  C [1x1] Regularization constant.
  solver [string] Solver; options are 'mdm','kozinec' or 'npa'.
  tmax [1x1] Maximal number of iterations.
  tolabs [1x1] Absolute tolerance stopping condition.
  tolrel [1x1] Relaitve tolerance stopping condition.

 Output:
  Alpha [nclass x num_data] Weights.
  bias [1x1] Bias.
  exitflag [1x1] Indicates which stopping condition was used:
    UB <= tolabs              ->  exit_flag = 1   Abs. tolerance.
    (UB-LB)/(LB+1) <= tolrel  ->  exit_flag = 2   Relative tolerance.
    t >= tmax                 ->  exit_flag = 0   Number of iterations.
  kercnt [1x1] Number of kernel evaluations.
  trnerr [1x1] Training error.
  t [1x1] Number of iterations.
  UB [1x1] Upper bound on the optimal solution.
  LB [1x1] Lower bound on the optimal solution.
  History [2x(t+1)] UB and LB with respect to number of iterations.

 Modifications:
 31-may-2004, VF
 25-jan-2003, VF
 24-jan-2003, VF
 23-jan-2003, VF
-------------------------------------------------------------------- */

#include "mex.h"
#include "matrix.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "kernel_fun.h"
#include "kozinec.h"
#include "mdm.h"
#include "npa.h"

#define INDEX(ROW,COL,DIM) ((COL*DIM)+ROW)

#define MINUS_INF INT_MIN
#define PLUS_INF  INT_MAX

#define KDELTA(A,B) (A==B)
#define KDELTA4(A1,A2,A3,A4) ((A1==A2) || (A1==A3) || (A1==A4) || (A2==A3) || (A2==A4) || (A3==A4))


long num_classes;   
double *labels;
double kernel_diag;

/* ----------------------------------------------- */
void get_indices2( long *index, long *class, long i )
{
   *index = i / (num_classes-1);
 
   *class = (i % (num_classes-1))+1;
   if( *class >= labels[ *index ]) (*class)++;

   return;
}

/* ------------------------------------------------------------
  Multi-class kernel function 
------------------------------------------------------------ */
double kernel_fce( long a, long b )
{
  double value;
  long i1,c1,i2,c2;

  get_indices2( &i1, &c1, a );
  get_indices2( &i2, &c2, b );

  if( KDELTA4(labels[i1],labels[i2],c1,c2) ) {
    value = (+KDELTA(labels[i1],labels[i2]) 
             -KDELTA(labels[i1],c2)
             -KDELTA(labels[i2],c1)
             +KDELTA(c1,c2)
            )*(kernel( i1, i2 )+1);
  }
  else
  {
    value = 0;
  }

  if(a==b) value += kernel_diag; 

  return( value );
}


/* ==============================================================
 Main MEX function - interface to Matlab.
============================================================== */
void mexFunction( int nlhs, mxArray *plhs[],
		  int nrhs, const mxArray*prhs[] )
{
  char solver[20];
  int exitflag;
  int *err_bit;
  int buf_len;
  long i,j ;
  long inx1, inx2;
  long num_data;
  long num_virt_data;
  long tmax;
  long t;
  double *tmp_ptr, *tmp_ptr1, *tmp_ptr2;
  double C;
  double tolrel;
  double tolabs;
  double LB;
  double UB;
  double *Alpha;
  double *History;
  double trnerr;


  /* == Input arguments == */
  if( nrhs != 9) mexErrMsgTxt("Incorrect number of input arguments.");

  dataA = mxGetPr(prhs[0]);   /* pointer at patterns */
  dataB = dataA;
  dim = mxGetM(prhs[0]);      /* data dimension */
  num_data = mxGetN(prhs[0]); /* number of data */
  labels = mxGetPr(prhs[1]);
  ker = kernel_id( prhs[2] );
  if( ker == -1 )
     mexErrMsgTxt("Improper kernel identifier.");
  arg1 = mxGetPr(prhs[3]);
  C = mxGetScalar(prhs[4]);

  /* take solver string */
  if( mxIsChar( prhs[5] ) != 1) 
     mexErrMsgTxt("solver must be string.");
  buf_len = (mxGetM(prhs[5]) * mxGetN(prhs[5])) + 1;
  buf_len = (buf_len > 20) ? 20 : buf_len;
  mxGetString( prhs[5], solver, buf_len );

  tmax = mxIsInf( mxGetScalar(prhs[6])) ? INT_MAX : (long)mxGetScalar(prhs[6]); 
  tolabs = mxGetScalar(prhs[7]);
  tolrel = mxGetScalar(prhs[8]);


  /* == Inicialization == */

  ker_cnt = 0;   

  if( C!=0 ) kernel_diag = 1/(2*C); else kernel_diag = 0;

  /* computes number of classes, i.e. max( labels ) */
  num_classes = MINUS_INF; 
  for( i = 0; i < num_data; i++ ) { 
     if( labels[i] > num_classes ) num_classes = (long)labels[i]; 
  }

  /* number of virtual patterns */
  num_virt_data = (num_classes-1)*num_data;

  /*  mexPrintf("num_classes=%d\n", num_classes);
  mexPrintf("num_virt_data=%d\n", num_virt_data);
  mexPrintf("num_data=%d\n", num_data);
  mexPrintf("C = %f, kernel_diag = %f\n", C, kernel_diag);*/

  /* == call SVM solver  == */
  if( strcmp( solver, "kozinec" )==0 ) {
    exitflag = single_kozinec( &kernel_fce, num_virt_data, tmax, tolabs, tolrel,
                               &Alpha, &UB, &LB, &t, &History );
  } else if ( strcmp( solver, "npa" )==0 ) {
    exitflag = single_npa( &kernel_fce, num_virt_data, tmax, tolabs, tolrel,
                            &Alpha, &UB, &LB, &t, &History );
  } else if ( strcmp( solver, "mdm" )==0 ) {
    exitflag = single_mdm( &kernel_fce, num_virt_data, tmax, tolabs, tolrel,
                           &Alpha, &UB, &LB, &t, &History );
  } else {
     mexErrMsgTxt("Unknown solver identifier.");
  }

  /*  mexPrintf("exitflag=%d\n", exitflag);
  mexPrintf("UB=%f, LB=%f\n", UB, LB);
  mexPrintf("t=%d\n", t );*/

  /* == Outputs == */
  plhs[0] = mxCreateDoubleMatrix(num_classes,num_data,mxREAL);
  tmp_ptr1 = mxGetPr(plhs[0]);

  plhs[1] = mxCreateDoubleMatrix(num_classes,1,mxREAL);
  tmp_ptr2 = mxGetPr(plhs[1]);

  for( i=0; i < num_classes; i++ ) {
    for( j=0; j < num_virt_data; j++ ) {
       get_indices2( &inx1, &inx2, j );

       tmp_ptr1[(inx1*num_classes)+i] += 
            Alpha[j]*(KDELTA(labels[inx1],i+1)+KDELTA(i+1,inx2));
       tmp_ptr2[i] += Alpha[j]*(KDELTA(labels[inx1],i+1)-KDELTA(i+1,inx2));
    }
  }

  /* */
  plhs[2] = mxCreateDoubleMatrix(1,1,mxREAL);
  *(mxGetPr(plhs[2])) = (double)exitflag;
  plhs[3] = mxCreateDoubleMatrix(1,1,mxREAL);
  *(mxGetPr(plhs[3])) = (double)ker_cnt;

  /* compute training error */
  err_bit = mxCalloc(num_data, sizeof(int));
  if( err_bit == NULL ) mexErrMsgTxt("Not enough memory.");
  for( i=0; i < num_classes; i++ ) {
    for( j=0; j < num_virt_data; j++ ) {
       get_indices2( &inx1, &inx2, j );
       if( Alpha[j] > 2*C ) err_bit[inx1] = 1; 
    }
  }

  for( trnerr = 0, i = 0; i < num_data; i++ ) trnerr += err_bit[i];

  /*  mexPrintf("# trn_er =% d\n", (long)trnerr); */

  trnerr = trnerr/num_data;
  plhs[4] = mxCreateDoubleMatrix(1,1,mxREAL);
  *(mxGetPr(plhs[4])) = trnerr;

  /* */
  plhs[5] = mxCreateDoubleMatrix(1,1,mxREAL);
  *(mxGetPr(plhs[5])) = (double)t;
  plhs[6] = mxCreateDoubleMatrix(1,1,mxREAL);
  *(mxGetPr(plhs[6])) = UB;
  plhs[7] = mxCreateDoubleMatrix(1,1,mxREAL);
  *(mxGetPr(plhs[7])) = LB;

  /* */
  plhs[8] = mxCreateDoubleMatrix(2,t+1,mxREAL);
  tmp_ptr = mxGetPr( plhs[8] );
  for( i = 0; i <= t; i++ ) {
     tmp_ptr[INDEX(0,i,2)] = History[INDEX(0,i,2)];
     tmp_ptr[INDEX(1,i,2)] = History[INDEX(1,i,2)];
  }

  /* */
  mxFree( Alpha );
  mxFree( History );

}

